/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmlp.h,v $
 * Revision 1.13  1995/05/16 09:39:34  adam
 * LICENSE.
 *
 * Revision 1.12  1995/03/02  08:06:05  adam
 * Fml function strsub implemented. New test files marc[45].fml.
 * New test options in fmltest.
 *
 * Revision 1.11  1995/02/27  09:01:21  adam
 * Regular expression support. Argument passing by name option. New FML
 * function strlen.
 *
 * Revision 1.10  1995/02/23  08:32:06  adam
 * Changed header.
 *
 * Revision 1.8  1995/02/10  18:15:52  adam
 * FML function 'strcmp' implemented. This function can be used to
 * test for existence of MARC fields.
 *
 * Revision 1.7  1995/02/10  15:50:56  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 * Revision 1.6  1995/02/09  16:06:07  adam
 * FML can be called from the outside multiple times by the functions:
 * fml_exec_call and fml_exec_call_str.
 * An interactive parameter (-i) to fmltest starts a shell-like
 * interface to FML by using the fml_exec_call_str function.
 *
 * Revision 1.5  1995/02/09  14:37:19  adam
 * Removed .depend from cvs. Removed function fml_mk_list.
 *
 * Revision 1.4  1995/02/09  14:33:37  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 * Revision 1.3  1995/02/09  13:07:15  adam
 * Nodes are freed now. Many bugs fixed.
 *
 * Revision 1.2  1995/02/07  16:09:23  adam
 * The \ character is no longer INCLUDED when terminating a token.
 * Major changes in tokenization routines. Bug fixes in expressions
 * with lists (fml_sub0).
 *
 * Revision 1.1.1.1  1995/02/06  13:48:10  adam
 * First version of the FML interpreter. It's slow and memory isn't
 * freed properly. In particular, the FML nodes aren't released yet.
 *
 */

#include <fml.h>

#ifndef FMLP_H
#define FMLP_H

#define FML_MAX_TOKEN 2048

#define FML_ATOM_BUF 12

struct fml_node {
    void *p[2];
    unsigned is_atom : 1;
};

struct fml_atom {
    struct fml_atom *next;
    char buf[FML_ATOM_BUF];
};

struct fml_node *fml_tokenize (Fml fml);
struct fml_node *fml_node_alloc (Fml fml);
struct fml_atom *fml_atom_alloc (Fml fml, char *str);
int fml_atom_str (struct fml_atom *a, char *str);
int fml_atom_len (struct fml_atom *a);
void fml_atom_strx (struct fml_atom *a, char *str, int max);
int fml_atom_val (struct fml_atom *a);
void fml_node_delete (Fml fml, struct fml_node *fn);
struct fml_node *fml_node_copy (Fml fml, struct fml_node *fn);
struct fml_node *fml_mk_node_val (Fml fml, int val);
int fml_atom_cmp (Fml fml, struct fml_atom *a1, struct fml_atom *a2);
struct fml_atom *fml_atom_strsub (Fml fml, struct fml_atom *a, int o, int l);

struct token {
    int             kind;
    int             separate;
    int             maxbuf;
    int             offset;
    char            *atombuf;
    char            *tokenbuf;

    int             escape_char;

    struct          fml_node *sub;
    struct          fml_atom *atom;
    char            sbuf[FML_ATOM_BUF*4];
};

struct fml_sym_info {
    int kind;
    struct fml_node *args;
    struct fml_node *body;
    struct fml_node *(*binary)(Fml fml, struct fml_node *l,
                               struct fml_node *r);
    struct fml_node *(*prefix)(Fml fml, struct fml_node **lp,
                               struct token *tp);
};

struct fml_sym_tab *fml_sym_open (void);
void fml_sym_close (struct fml_sym_tab **tabp);
struct fml_sym_info *fml_sym_add (struct fml_sym_tab *tab, const char *s);
struct fml_sym_info *fml_sym_add_local (struct fml_sym_tab *tab,const char *s);

struct fml_sym_info *fml_sym_lookup (struct fml_sym_tab *tab, const char *s);
struct fml_sym_info *fml_sym_lookup_local (struct fml_sym_tab *tab,
                                           const char *s);
void fml_sym_push (struct fml_sym_tab *tab);
void fml_sym_pop (struct fml_sym_tab *tab, void(*ph)(struct fml_sym_info *i));

void fml_pr_list (struct fml_node *p);
void fml_node_stat (Fml fml);

#define FML_FUNC     1
#define FML_IF       2
#define FML_ELSE     3
#define FML_PREFIX   4
#define FML_VAR      5
#define FML_FOREACH  6
#define FML_RETURN   7
#define FML_SET      8
#define FML_WHILE    9
#define FML_CBINARY 10
#define FML_CPREFIX 11
#define FML_BINARY  12
#define FML_BIN     13
#define FML_CODE    14

void fml_rel_init (Fml fml);
void fml_arit_init (Fml fml);
void fml_list_init (Fml fml);
void fml_str_init (Fml fml);
void fml_lr_values (Fml fml, struct fml_node *l, int *left_val,
                           struct fml_node *r, int *right_val);
void fml_cmd_lex (struct fml_node **np, struct token *tp);
void fml_cmd_lex_s (struct fml_node **np, struct token *tp, int esc_stop);
void fml_init_token (struct token *tp, Fml fml);
void fml_del_token (struct token *tp, Fml fml);
struct fml_node *fml_expr_term (Fml fml, struct fml_node **lp, 
                                struct token *tp);
struct fml_node *fml_exec_group (struct fml_node *list, Fml fml);
#endif
