/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Log: wcgi.c,v $
 * Revision 1.18  1996/03/14 11:48:37  adam
 * New function egw_prog that returns name of shell.
 *
 * Revision 1.17  1996/02/26  10:36:15  adam
 * Better error handling when (re)spawn of the shell fails.
 *
 * Revision 1.16  1996/02/12  10:10:29  adam
 * Resource/config system used by the gateway.
 *
 * Revision 1.15  1996/01/26  09:02:20  adam
 * Open of client FIFO called with O_NDELAY when reconnecting to shell
 * in order to prevent serious lock if previous shell died without
 * unlinking client FIFO.
 *
 * Revision 1.14  1996/01/12  13:08:06  adam
 * CGI script passes name of lock file to the shell. The server will not close
 * the response FIFO until this file becomes unlocked. This method handles
 * cancel operations much better.
 *
 * Revision 1.13  1996/01/12  10:05:17  adam
 * If script name ends with ';' HTTP/GET/Expires will be defined.
 * The cgi interface only reads final handshake if response from
 * server (shell) was zero-terminated [If it isn't it probably died].
 *
 * Revision 1.12  1996/01/09  10:46:49  adam
 * New defines: LOGDIR/EGWDIR/CGIDIR set in Makefile.
 *
 * Revision 1.11  1996/01/08  08:42:19  adam
 * Handles method GET.
 *
 * Revision 1.10  1996/01/05  16:21:20  adam
 * Bug fix: shell (wproto) sometimes closed server FIFO before cgi
 * program opened it - solution: cgi sends OK when response has been read.
 *
 * Revision 1.9  1995/12/20  16:31:33  adam
 * Bug fix: shell might terminate even though new request was initiated
 * by the cgi interface program.
 * Work on more simple user interface and Europagate buttons.
 *
 * Revision 1.8  1995/11/08  16:14:35  adam
 * Many improvements and bug fixes.
 * First version that ran on dtbsun.
 *
 * Revision 1.7  1995/11/08  12:42:18  adam
 * Added descriptive text field in target info.
 * Added authentication field in target info.
 *
 * Revision 1.6  1995/11/06  17:44:22  adam
 * State reestablised when shell restarts. History of previous
 * result sets.
 *
 * Revision 1.5  1995/11/06  10:51:15  adam
 * End of response marker in response from wsh/wproto to wcgi.
 * Shells are respawned when necessary.
 *
 * Revision 1.4  1995/11/02  16:35:37  adam
 * Bug fixes and select on FIFOs in wcgi - doesn't really work!
 *
 * Revision 1.3  1995/10/31  16:56:24  adam
 * Record presentation.
 *
 * Revision 1.2  1995/10/23  16:55:36  adam
 * A lot of changes - really.
 *
 * Revision 1.1  1995/10/20  11:49:25  adam
 * First version of www gateway.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#ifdef AIX
#include <sys/select.h>
#endif

#include <gw-db.h>
#include <gw-res.h>
#include "wproto.h"

static char *prog = "cgi";

static char serverp[256] = {'\0'};
static char serverf[256] = {'\0'};
static GW_DB gw_db = NULL;

static void fatal(char *p)
{
    printf("Content-type: text/html\n\n<HTML><HEAD><TITLE>Server Failure</TITLE></HEAD>\n");
    printf("<BODY>%s</BODY>\n", p);
    if (gw_db)
        gw_db_close (gw_db);
    if (*serverp)
    	unlink (serverp);
    if (*serverf)
        unlink (serverf);
    exit(0);
}

static int spawn (char *sprog, int id)
{
    int r, fd;
    char path[256];
    char envstr[80];

    sprintf (envstr, "GWID=%d", id);
    putenv (envstr);
    sprintf(path, "%s/%s", EGWDIR, sprog);
    switch(r = fork())
    {
    	case -1: 
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "fork"); 
            fatal ("Internal error in server");
    	case 0: 
            close (0);
            close (1);
	    close (2);
            gw_log (GW_LOG_DEBUG, prog, "execl %s", path);
            execl (path, sprog, 0); 
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "execl %s", path);
            fd = open (serverp, O_WRONLY);
            if (fd >= 0)
            {
                write (fd, "FA", 2);
                close (fd);
            }
	    exit(1);
    	default: 
            return r;
    }
}

#if 0
static void print_environ (void)
{
    extern char **environ;
    int i;

    for (i = 0; environ[i]; i++)
        gw_log (GW_LOG_DEBUG, prog, "e: %s", environ[i]);
}
#endif

/*
 * NOTE: In the (perhaps odd) terminology used within this software,
 * the 'server' is the present program, which is executed by the httpd
 * server. The 'client' is the process running outside.
 * Protocol is long(len)<serverfifo>\0<extrapath>\0<envvars>\0<INFO>\0
 */
int main()
{
    char clientp[256], *path_info, *p, *operation, *t;
    char combuf[COMBUF];
    const char *fifoDir;
    GwRes cgiRes;
    int serverf_fd = -1;
    int linein = -1, lineout, data, gw_id;

    if (chdir (EGWDIR))
        fatal ("Couldn't change directory to " EGWDIR);
    gw_log_init ("egw");
    gw_log_file (GW_LOG_ALL, "egwcgi_log");
#if 0
    gw_log_level (GW_LOG_ALL);
#endif
    gw_log_session (getpid());
    gw_log (GW_LOG_STAT, prog, "Europagate www cgi server");
    cgiRes = gw_res_init ();
    gw_res_merge (cgiRes, "egw.res");

#if 1
    gw_log_level (gw_log_mask_str (
                  gw_res_get (cgiRes, "log.level", "default")));
#endif
    fifoDir = gw_res_get (cgiRes, "fifo.dir", "/tmp/egw");
    
    /* Create fifo directory if it doesn't exist already */
    if (access(fifoDir, R_OK|W_OK) < 0 && mkdir(fifoDir, 0777) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "Failed to create %s",
                fifoDir);
    	fatal("Internal error in server.");
    }
    /* Delete server FIFO if it does exist */
    sprintf(serverp, "%s/srv%d", fifoDir, getpid());
    if (access(serverp, R_OK|W_OK) == 0)
    {
    	if (unlink(serverp) < 0)
    	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog,
                    "Failed to delete stale fifo.");
	    fatal("Internal error in server.");
	}
	else
	    gw_log (GW_LOG_WARN, prog, "Removed stale server fifo.");
    }
    /* Make server FIFO */
    if (mkfifo(serverp, 0666 | S_IFIFO) < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "mkfifo(%s)", serverp);
	fatal("Internal error in server.");
    }
    /* The httpd server must pass PATH_INFO */
    if (!(path_info = getenv("PATH_INFO")))
    {
    	gw_log (GW_LOG_FATAL, prog, "Must set PATH_INFO.");
    	fatal("Internal error in server.");
    }
    /* Create lock file that ensures the server (shell) doesn't */
    /* terminate before we have read the whole response */
    sprintf (serverf, "%s/srf%d", fifoDir, getpid ());
    gw_log (GW_LOG_DEBUG, prog, "open w %s", serverf);
    serverf_fd = open (serverf, O_WRONLY|O_CREAT|O_TRUNC, 0666);
    if (serverf_fd == -1)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "open (%s)", serverf);
	fatal("Internal error in server.");
    }
    else
    {
        struct flock area;
        area.l_type = F_WRLCK;
        area.l_whence = SEEK_SET;
        area.l_start = 0L;
        area.l_len = 0L;
        fcntl (serverf_fd, F_SETLK, &area);
    }
    /* Read first part of path_info. Either it is a numeric session id */
    /* or it is the name of a backend shell. */
    operation = ++path_info;
    while (*path_info && *path_info != '/')
    	path_info++;
    if (*path_info)
    	*(path_info++) = '\0';
    gw_log (GW_LOG_DEBUG, prog, "www.db open");
    if (!(gw_db = gw_db_open (EGWDIR "/www.db", 1, 1)))
    {
        gw_log (GW_LOG_FATAL, prog, "gw_db_open");
        exit (1);
    }
    gw_log (GW_LOG_DEBUG, prog, "www.db ok");
    /* Is operation a backend shell (new session) ? */
    if ((gw_id = atoi(operation)) <= 0)
    {
        int r;
        char gw_id_str[16];

        /* Get new unique id */
        gw_id = gw_db_seq_no (gw_db);
        sprintf (gw_id_str, "%d", gw_id);
       
        /* Spawn backend shell (server) */ 
    	spawn(operation, gw_id);
        r = gw_db_insert (gw_db, gw_id_str, strlen(gw_id_str)+1,
                          operation, strlen(operation)+1);
        if (r)
        {
            gw_log (GW_LOG_FATAL, prog, "gw_db_insert: %d", r);
            gw_db_close (gw_db);
            exit (1);
        }
    	gw_log (GW_LOG_DEBUG, prog, "Synchronizing with client");
	if ((linein = open(serverp, O_RDONLY)) < 0)
	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "open r %s", serverp);
	    fatal("Internal error in server.");
	}
	if (read(linein, combuf, 2) < 2 || strcmp(combuf, "OK"))
	{
	    gw_log (GW_LOG_FATAL, prog, "Failed to synchronize with client");
	    fatal("Internal error in server");
	}
    	gw_log (GW_LOG_DEBUG, prog, "Synchronized");
        sprintf(clientp, "%s/clt%d", fifoDir, gw_id);
        gw_log (GW_LOG_DEBUG, prog, "open w %s", clientp);
        lineout = open (clientp, O_WRONLY);
    }
    else /* A session is continued */
    {
        sprintf(clientp, "%s/clt%d", fifoDir, gw_id);
        gw_log (GW_LOG_DEBUG, prog, "open w|n %s", clientp);
        /* Open the FIFO in O_NDELAY-mode: This prevents blocking */
        /* even though the shell died without unlinking the FIFO */
        /* On the other hand, if the shell is running, it will never */
        /* close this FIFO */
        lineout = open (clientp, O_WRONLY|O_NDELAY);
    }
    /* If open of clientp failed, the shell is not running, so we */
    /* invoke it again */
    if (lineout < 0)
    {
        char gw_id_str[16];
        void *sprog;
        size_t sprog_size;
        int r;

        sprintf (gw_id_str, "%d", gw_id);
        r = gw_db_lookup (gw_db, gw_id_str, strlen(gw_id_str)+1,  
                          &sprog, &sprog_size);
        if (r != 1)
        {
	    gw_log (GW_LOG_FATAL, prog, "gw_db_lookup %s", gw_id_str);
	    fatal("Internal error in server");
        }
    	gw_log (GW_LOG_DEBUG|GW_LOG_ERRNO, prog, "open r %s restart", clientp);
    	spawn (sprog, gw_id);
    	gw_log (GW_LOG_DEBUG, prog, "Synchronizing with client");
	if ((linein = open(serverp, O_RDONLY)) < 0)
	{
	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "open %s", serverp);
	    fatal("Internal error in server");
	}
	if (read(linein, combuf, 2) < 2 || strcmp(combuf, "OK"))
	{
	    gw_log (GW_LOG_FATAL, prog, "Failed to synchronize with client.");
	    fatal("Internal error in server");
	}
    	gw_log (GW_LOG_DEBUG, prog, "Synchronized.");
        if ((lineout = open(clientp, O_WRONLY)) < 0)
        {
    	    gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "%s", clientp);
	    fatal("Internal error in server");
        }
    }
    gw_db_close (gw_db);
    gw_log (GW_LOG_DEBUG, prog, "Decoding user data");
    p = combuf + sizeof(data);
    strcpy (p, serverp);
    p += strlen (p) + 1;
    strcpy (p, serverf);
    p += strlen (p) + 1;
    strcpy (p, path_info);
    gw_log (GW_LOG_STAT, prog, "P:%s", p);
    p += strlen(p) + 1;
    *(p++) = '\0';               /* no envvars tranferred at present */
    if ((t = getenv("CONTENT_LENGTH")) && (data = atoi(t)) > 0)
    {
        int j, i = 0;
        while (i < data)
        {
     	    j = read(0, p + i, data - i);
            if (j == -1)
            {
	        gw_log (GW_LOG_ERRNO|GW_LOG_FATAL, prog,
                        "Failed to read input");
	        fatal("Internal error in server");
            }
            else if (j == 0)
            {
	        gw_log (GW_LOG_ERRNO, prog, "Failed to read input");
	        fatal("Internal error in server");
            }
            i += j;
	}
    }
    else if ((t = getenv("QUERY_STRING")))
    {
        strcpy (p, t);
        data = strlen(p);
    }
    p[data] = '\0';
    gw_log (GW_LOG_DEBUG, prog, "C:%s", p);
    p += data+1;
    data = (p - combuf);
    memcpy(combuf, &data, sizeof(data));
    gw_log (GW_LOG_DEBUG, prog, "Writing data");
    if (write(lineout, combuf, data) < data)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "write");
    	fatal("Internal server error");
    }
    if (linein < 0)
    {
        gw_log (GW_LOG_DEBUG, prog, "open r %s", serverp);
        if ((linein = open(serverp, O_RDONLY)) < 0)
        {
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "open %s", serverp);
            fatal("Internal error in server");
        }
    }
    gw_log (GW_LOG_DEBUG, prog, "Reading response");

    while ((data = read(linein, combuf, COMBUF)) > 0)
    {
        gw_log (GW_LOG_DEBUG, prog, "Got %d bytes", data);
	if (combuf[data-1] == '\0')
            break;
        if (write(1, combuf, data) < data)
        {
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "write");
            exit (1);
        }
    }
    if (data < 0)
    {
    	gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "read");
        exit (1);
    }
    if (data > 0)
    {
        if (close (serverf_fd))
        {
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "close %s", serverf);
        }
        if (--data > 0)
        {
            if (write(1, combuf, data) < data)
            {
                gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, prog, "write");
                exit (1);
            }
        }
    }
    gw_log (GW_LOG_DEBUG, prog, "Cleaning up.");
    unlink (serverf);
    close(linein);
    unlink(serverp);
    close(lineout);
    return 0;
}
