/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Log: wproto.h,v $
 * Revision 1.11  1996/05/23 15:53:11  adam
 * Bug fix: egw_enc failed on 8-bit chars.
 * New command: egw_parms.
 *
 * Revision 1.10  1996/05/21  14:53:04  adam
 * Tcl command wform extented; options -raw and -exists added.
 *
 * Revision 1.9  1996/03/14  11:48:39  adam
 * New function egw_prog that returns name of shell.
 *
 * Revision 1.8  1996/02/12  10:09:25  adam
 * New parameter to wproto_init: directory root for the FIFOs (instead
 * of using preprocessor defines FIFODIR/FIFOROOT).
 *
 * Revision 1.7  1996/01/24  08:26:55  adam
 * All tcl commands prefixed with egw_ (except the html command).
 *
 * Revision 1.6  1996/01/12  13:08:07  adam
 * CGI script passes name of lock file to the shell. The server will not close
 * the response FIFO until this file becomes unlocked. This method handles
 * cancel operations much better.
 *
 * Revision 1.5  1996/01/12  10:05:21  adam
 * If script name ends with ';' HTTP/GET/Expires will be defined.
 * The cgi interface only reads final handshake if response from
 * server (shell) was zero-terminated [If it isn't it probably died].
 *
 * Revision 1.4  1995/11/06  10:51:18  adam
 * End of response marker in response from wsh/wproto to wcgi.
 * Shells are respawned when necessary.
 *
 * Revision 1.3  1995/10/27  15:12:11  adam
 * IrTcl incorporated in the gateway.
 * Better separation of script types.
 * Z39.50 gateway scripts entered.
 *
 * Revision 1.2  1995/10/23  16:55:41  adam
 * A lot of changes - really.
 *
 * Revision 1.1  1995/10/20  11:49:27  adam
 * First version of www gateway.
 *
 */

#ifndef WPROTO_H
#define WPROTO_H

#include <gw-log.h>
#include <gw-res.h>

#define COMBUF 4096
#define OUTBUFFER_CHUNK 4096

typedef struct wform_data
{
    char name[64];
    char value[512];
} wform_data;

typedef struct cache_data
{
    char name[256];
    char path[256];
} cache_data;

struct w_select_handle {
    int  fd;
    struct {
        void (*sproc)(int, void *);
        void *obj;
    } rwx[3];
    struct w_select_handle *next;
};

typedef struct wclient_data
{
    int id;
    char *prog;
    char *raw_data;
    wform_data wf_data[50];
    wform_data wf_parms_var[50];
    char wf_parms[1024];
    char wf_serverp[1024];
    char wf_serverf[1024];
    char wf_script[1024];
    char path[512];
    int linein;
    int lineout;
    char *outbuffer;
    int outbuffer_size;
    int outbuffer_offset;
    int cache_level;
    cache_data cache[20];
    int cache_fd;
    struct w_select_handle *select_list;
    const char *fifoDir;
} *WCLIENT, wclient_data;

#define wo_putc(wc, ch)   \
    ((wc)->outbuffer_offset >= OUTBUFFER_CHUNK ? wo_overflow((wc), (ch)) :  \
    (*((wc)->outbuffer + (wc)->outbuffer_offset++) = (char) (ch)), 0)

char *wgetval (WCLIENT wc, char *name);
int wproto_process (WCLIENT wc, int timeout);
WCLIENT wproto_init (const char *fifoDir, const char *prog);
void wo_printf (WCLIENT wc, const char *fmt, ...);
void wo_clear (WCLIENT wc, const char *type);
int wo_finish (WCLIENT wc);
int wo_flush (WCLIENT wc);
int wo_overflow (WCLIENT wc, char ch);
void wproto_terminate (WCLIENT wc);
int wproto_cache (WCLIENT wc, int level);
int wo_puthtml (WCLIENT wc, char *name);
void wo_puts (WCLIENT wc, const char *s);
void wo_write (WCLIENT wc, const char *s, size_t len);

#endif
